// Do not edit this file, it is autogenerated by CMake from the .md file
_HKI( "<table>\n"
"    <tr>\n"
"        <th>Markup</th>\n"
"        <th></th>\n"
"        <th>Result</th>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>^{superscript}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp><sup>superscript</sup>&nbsp;</samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>Driver Board^{Rev A}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp>Driver Board<sup>Rev A</sup></samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td><br></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>_{subscript}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp><sub>subscript</sub>&nbsp;</samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>D_{0} - D_{15}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp>D<sub>0</sub> - D<sub>15</sub></samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td> &nbsp;<br><samp>~{overbar}</samp><br> &nbsp;<br><samp>~{CLK}</samp></td>\n"
"        <td></td>\n"
"        <td> <samp><u>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</u></samp><br> <samp>overbar</samp><br> <samp><u>&nbsp;&nbsp;&nbsp;</u></samp><br> <samp>CLK</samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>${variable}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp><i>variable_value</i></samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>${REVISION}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp>2020.1</samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td><br></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>${refdes:field}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp><i>field_value</i> of symbol <i>refdes</i></samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>${R3:VALUE}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp>150K</samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td><br></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td><br></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <th>Bus Definition</th>\n"
"        <th>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</th>\n"
"        <th>Resultant Nets</th>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>prefix[m..n]</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp>prefixm to prefixn</samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>D[0..7]</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp>D0, D1, D2, D3, D4, D5, D6, D7</samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td><br></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>{net1 net2 ...}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp>net1, net2, ...</samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>{SCL SDA}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp>SCL, SDA</samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td><br></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>prefix{net1 net2 ...}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp>prefix.net1, prefix.net2, ...</samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>USB1{D+ D-}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br><samp>USB1.D+, USB1.D-</samp></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td><br></td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>MEM{D[1..2] LATCH}</samp></td>\n"
"        <td></td>\n"
"        <td>&nbsp;<br> <samp>MEM.D1, MEM.D2, MEM.LATCH</samp> </td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td>&nbsp;<br><samp>MEM{D_{[1..2]} ~{LATCH}}</samp></td>\n"
"        <td></td>\n"
"        <td> <samp>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<sub>&nbsp;</sub>&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<sub>&nbsp;</sub>&nbsp; &nbsp;&nbsp;&nbsp;&nbsp;<u>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</u></samp><br> <samp>MEM.D<sub>1</sub>, MEM.D<sub>2</sub>, MEM.LATCH</samp> </td>\n"
"    </tr>\n"
"    <tr>\n"
"        <td><br></td>\n"
"    </tr>\n"
"</table>\n"
"<p></p>\n"
"<p></p>\n"
"<i>Note that markup has precedence over bus definitions.</i>\n"
"\n"
"" );
