/*
 * This program source code file is part of KiCad, a free EDA CAD application.
 *
 * Copyright (C) 2021 Mikołaj Wielgus <wielgusmikolaj@gmail.com>
 * Copyright The KiCad Developers, see AUTHORS.txt for contributors.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you may find one here:
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 * or you may search the http://www.gnu.org website for the version 2 license,
 * or you may write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

/**
 * This file contains the file format version information for the s-expression drawing sheet file
 * formats
 *
 * @note Comment out the last version and add the new version as a date time stamp in the YYYYMMDD
 * format.  Comment the changes to the file format for historical purposes.
 */

//#define SEXPR_WORKSHEET_FILE_VERSION 20210606 // Initial version.
//#define SEXPR_WORKSHEET_FILE_VERSION 20220228 // Font support.
//#define SEXPR_WORKSHEET_FILE_VERSION 20230607 // Save images as base64.
#define SEXPR_WORKSHEET_FILE_VERSION   20231118 // Generator version; V8 file format cleanup
