#include "resource.h"

IDI_APP_PCBCALC_ICON ICON "../resources/bitmaps_png/icons/icon_pcbcalculator.ico"

#define RC_VER_FILE_DESCRIPTION "KiCad PCB Calculator " KICAD_WIN32_RC_PRODVER_STR
#define RC_VER_INTERNALNAME "pcb_calculator"
#define RC_VER_ORIGINALFILENAME "pcb_calculator.exe"
#include "kiwin32.rc"
