// Do not edit this file, it is autogenerated by CMake from the .md file
_HKI( "### Notes\n"
      "\n"
      "Use Ctrl+/ to comment or uncomment line(s).\n"
      "<br><br><br>\n"
      "\n"
      "" );
